#include "eval.h"
#include "program.h"
#include "macro.h"
#include "standard.h"
#include "primitive.h"

struct scm_thing * c_eval(struct scm_thing * expression,
			  struct scm_thing * local_environment,
			  struct scm_thing * top_level_environment) {
  if (!expression) {
    // can't eval empty list!
    error();
  }

  switch (expression->type) {
  case PAIR:
    // eval the first member of the list
    struct scm_thing *function = c_eval(expression->car,
					local_environment,
					top_level_environment);
    switch (function->type) {
    case COMPILED_PROCEDURE:
    case INTERPRETED_PROCEDURE:
      return c_procedure_call(function,
			      expression->cdr,
			      local_environment,
			      top_level_environment);
    case COMPILED_MACRO:
      return (*function->macro)(expression->cdr,
				local_environment,
				top_level_environment);
    case INTERPRETED_MACRO:
      // this will be the cdr of a syntax-rules
      struct scm_thing * literals = function->car;
      struct scm_thing * syntax_rules = function->cdr;
      // we need to exclude the macro
      struct scm_thing * form = expression->cdr;
      while (syntax_rules) {
	// exclude the macro name
	struct scm_thing * pattern = syntax_rules->car->car->cdr;
	if (test_pattern(pattern,
			 literals,
			 form,
			 local_environment,
			 top_level_environment)) {
	  break;
	} else {
	  syntax_rules = syntax_rules->cdr;
	}
      }
      if (!syntax_rules) {
	// no match
	printf("no appropriate syntax match found\n");
	error();
      }
      // we use a special macro evaluator so we
      // don't leak as stupid of an amount of memory
      struct scm_thing * syntax_rule = syntax_rules->car;
      struct scm_thing * template = syntax_rule->cdr->car;
      struct scm_thing * matches = match_pattern(syntax_rule->car->cdr,
						 literals,
						 form,
						 NULL);
      struct scm_thing ** sterile = malloc(sizeof(struct scm_thing *));
      *sterile = NULL;
      struct scm_thing * transcribed = transcribe_syntax_template(template,
								  matches,
								  sterile,
								  top_level_environment);
      free_alist_structure(*sterile);
      free(sterile);
      // don't leak so much damn memory!
      /*
      while (matches) {
	struct scm_thing * next_matches = matches->cdr;
	free(matches->car);
	free(matches);
	matches = next_matches;
      }
      */
      return c_eval(transcribed,
		    local_environment,
		    top_level_environment);
      break;
    default:
      // not a function!!!
      if (expression->car->type == SYMBOL) {
	printf("%s is not a function!\n", expression->car->symbol);
      } else {
	printf("attempt to call nonsense?\n");
      }
      error();
      break;
    }
    break;
  case SYMBOL:
    struct scm_thing * pair = c_assoc(expression, local_environment);
    if (pair) {
      // maybe this is a bad idea, actually?
      // yeah symbols are evaled before being defined
      //return c_eval(variable_pair->cdr, variables);
      return pair->cdr;
    } else {
      pair = c_assoc(expression, top_level_environment);
      if (pair) {
	return pair->cdr;
      } else {
	// symbol but not variable?
	printf("unrecognised symbol: %s\n", expression->symbol);
	error();
      }
    }
    break;
  case INDIRECTION:
    return expression->car->cdr;
  default:
    // pretty sure everything else evals to itself
    return expression;
  }
}

struct scm_thing * c_evlis(struct scm_thing * list,
			   struct scm_thing * local_environment,
			   struct scm_thing * top_level_environment) {
  if (!list) {
    return NULL;
  }
  struct scm_thing *result_head = malloc(sizeof(struct scm_thing));
  struct scm_thing *result_tail = result_head;
  result_tail->type = PAIR;
  result_tail->car = c_eval(list->car,
			    local_environment,
			    top_level_environment);
  list = list->cdr;
  while (list) {
    result_tail->cdr = malloc(sizeof(struct scm_thing));
    result_tail = result_tail->cdr;
    result_tail->type = PAIR;
    result_tail->car = c_eval(list->car,
			      local_environment,
			      top_level_environment);
    list = list->cdr;
  }
  result_tail->cdr = NULL;
  return result_head;
}
