#include "macro.h"
#include "standard.h"

struct scm_thing * get_binding(struct scm_thing * identifier,
			       struct scm_thing * local_environment,
			       struct scm_thing * top_level_environment) {
  // assume identifier is an identifier
  struct scm_thing * pair = c_assoc(identifier, local_environment);
  if (pair) {
    return pair->cdr;
  } else {
    pair = c_assoc(identifier, top_level_environment);
    if (pair) {
      return pair->cdr;
    } else {
      return identifier;
    }
  }
}

void free_alist_structure(struct scm_thing * alist) {
  while (alist) {
    struct scm_thing * old_alist = alist;
    alist = alist->cdr;
    free(old_alist->car);
    free(old_alist);
  }
}

bool is_elided(struct scm_thing * list) {
  if (list->cdr &&
      list->cdr->type == PAIR &&
      c_equal_p(list->cdr->car, &scm_ellipsis)) {
    return true;
  } else {
    return false;
  }
}

struct scm_thing * new_pair(void) {
  struct scm_thing * pair = malloc(sizeof(struct scm_thing));
  pair->type = PAIR;
  pair->mutable = true;
  return pair;
}

struct scm_thing * append_pair(struct scm_thing * original) {
  struct scm_thing * pair = malloc(sizeof(struct scm_thing));
  pair->type = PAIR;
  pair->mutable = true;
  original->cdr = pair;
  return pair;
}

// does not work if you have a variable followed by multiple ellipses
struct scm_thing * find_all_identifiers(struct scm_thing * template,
					struct scm_thing * identifiers) {
  while (template && template->type == PAIR) {
    if (is_elided(template)) {
      struct scm_thing * new_identifiers = find_all_identifiers(template->car,
								NULL);
      struct scm_thing * slider = new_identifiers;
      while (slider) {
	identifiers = c_cons(c_cons(slider->car,
				    &scm_ellipsis_pair),
			     identifiers);
	struct scm_thing * old_slider = slider;
	slider = slider->cdr;
	free(old_slider);
      }
      template = template->cdr->cdr;
    } else {
      identifiers = find_all_identifiers(template->car,
					 identifiers);
      template = template->cdr;
    }
  }
  if (template && template->type == SYMBOL) {
    return c_cons(template, identifiers);
  } else {
    return identifiers;
  }
}

uint32_t count_ellipses(struct scm_thing * list) {
  uint32_t i = 0;
  while (list &&
	 list->type == PAIR &&
	 c_equal_p(list->car, &scm_ellipsis)) {
    i++;
    list = list->cdr;
  }
  return i;
}

struct scm_thing * transcribe_syntax_template_elided(struct scm_thing * template_big,
						     struct scm_thing * matches,
						     struct scm_thing ** sterile,
						     struct scm_thing * top_level_environment) {
  struct scm_thing * template = template_big->car;
  uint32_t ellipses = count_ellipses(template_big->cdr);
  if (ellipses != 1) {
    printf("too many ellipses\n");
    error();
  }
  struct scm_thing * remainder = template_big->cdr;
  while (remainder &&
	 c_equal_p(remainder->car, &scm_ellipsis)) {
    remainder = remainder->cdr;
  }
  if (template->type == SYMBOL) {
    if (remainder) {
      // may be buggy
      //printf("WARNING: running untested codepath\n");
      struct scm_thing * pair = c_assoc(c_cons(template,
					       &scm_ellipsis_pair),
					matches);
      if (pair && pair->cdr) {
	struct scm_thing * slider = pair->cdr;
	struct scm_thing * result_head = NULL;
	struct scm_thing * result_end = NULL;
	while (slider) {
	  if (!result_end) {
	    result_head = malloc(sizeof(struct scm_thing));
	    result_end = result_head;
	  } else {
	    result_end->cdr = malloc(sizeof(struct scm_thing));
	    result_end = result_end->cdr;
	  }
	  result_end->type = PAIR;
	  result_end->mutable = true;
	  result_end->car = slider->car;
	  slider = slider->cdr;
	}
	result_end->cdr = transcribe_syntax_template(remainder,
						     matches,
						     sterile,
						     top_level_environment);
	return result_head;
      } else {
	return transcribe_syntax_template(remainder,
					  matches,
					  sterile,
					  top_level_environment);
      }
    } else {
      struct scm_thing * pair = c_assoc(template_big, matches);
      if (pair) {
	return pair->cdr;
      } else {
	return NULL;
      }
    }
  } else if (template->type == PAIR) {
    // why is it called matcheses? because each value in the alist
    // is a list of all the matches of the identifier
    struct scm_thing * new_matcheses = NULL;
    struct scm_thing * slider = matches;
    // this should exclude identifiers not in the template
    struct scm_thing * identifiers = find_all_identifiers(template,
							  NULL);
    // will collect all of the matches that have the correct number of ellipses
    // and are in identifiers, then remove the ellipses
    while (slider) {
      if (count_ellipses(slider->car->car->cdr) == ellipses &&
	  c_member(slider->car->car->car, identifiers)) {
	new_matcheses = c_cons(c_cons(slider->car->car->car,
				      slider->car->cdr),
			       new_matcheses);
      }
      slider = slider->cdr;
    }
    // now match them up
    if (new_matcheses) {
      struct scm_thing * transcribed_head = NULL;
      struct scm_thing * transcribed_tail = transcribed_head;
      while (new_matcheses->car->cdr) {
	// do a confusing thing to isolate each instance
	struct scm_thing * new_matches = NULL;
	slider = new_matcheses;
	while (slider) {
	  new_matches = c_cons(c_cons(slider->car->car,
				      slider->car->cdr->car),
			       new_matches);
	  slider->car->cdr = slider->car->cdr->cdr;
	  slider = slider->cdr;
	}
	if (transcribed_tail) {
	  transcribed_tail = append_pair(transcribed_tail);
	} else {
	  transcribed_head = new_pair();
	  transcribed_tail = transcribed_head;
	}
	transcribed_tail->car = transcribe_syntax_template(template,
							   new_matches,
							   sterile,
							   top_level_environment);
	free_alist_structure(new_matches);
      }
      free_alist_structure(new_matcheses);
      transcribed_tail->cdr = transcribe_syntax_template(remainder,
							 matches,
							 sterile,
							 top_level_environment);
      return transcribed_head;
    } else {
      // there were no matches
      return transcribe_syntax_template(remainder,
					matches,
					sterile,
					top_level_environment);
    }
  } else {
    error();
  }
}

struct scm_thing * transcribe_syntax_template(struct scm_thing * template,
					      struct scm_thing * matches,
					      struct scm_thing ** sterile,
					      struct scm_thing * top_level_environment) {
  if (!matches) {
    printf("no matches: probably an error\n");
  }
  if (!template) {
    return NULL;
  } else if (template->type == PAIR) {
    if (is_elided(template)) {
      return transcribe_syntax_template_elided(template,
					       matches,
					       sterile,
					       top_level_environment);
    } else {
      struct scm_thing * result_head = NULL;
      struct scm_thing * result_tail = result_head;
      while (template &&
	     template->type == PAIR &&
	     !is_elided(template)) {
	if (result_tail) {
	  result_tail->cdr = malloc(sizeof(struct scm_thing));
	  result_tail = result_tail->cdr;
	} else {
	  result_head = malloc(sizeof(struct scm_thing));
	  result_tail = result_head;
	}
	result_tail->type = PAIR;
	result_tail->car = transcribe_syntax_template(template->car,
						      matches,
						      sterile,
						      top_level_environment);
	template = template->cdr;
      }
      if (result_tail) {
	result_tail->cdr = transcribe_syntax_template(template,
						      matches,
						      sterile,
						      top_level_environment);
	return result_head;
      } else {
	return transcribe_syntax_template(template,
					  matches,
					  sterile,
					  top_level_environment);
      }
    }
  } else if (template->type == VECTOR) {
    if (c_equal_p(template->vector[template->vectorsize - 1],
		  &scm_ellipsis)) {
      // IMPLEMENTATION RESTRICTION
      error();
    }
    struct scm_thing * new_vector = c_make_vector(template->vectorsize);
    for (uint32_t i = 0; i < template->vectorsize; i++) {
      new_vector->vector[i] = transcribe_syntax_template(template->vector[i],
							 matches,
							 sterile, top_level_environment);
    }
    return new_vector;
  } else if (template->type == SYMBOL) {
    struct scm_thing * pair = c_assoc(template, matches);
    if (pair) {
      return pair->cdr;
    } else {
      pair = c_assoc(template, top_level_environment);
      if (pair) {
	return pair->cdr;
      } else {
	// turns out macros basically have closure, who would have guessed?
	pair = c_assoc(template, *sterile);
	if (pair) {
	  return pair->cdr;
	} else {
	  // this is a binding, make up an indirection
	  struct scm_thing * indirection = malloc(sizeof(struct scm_thing));
	  indirection->type = INDIRECTION;
	  indirection->car = c_cons(template, &scm_undefined);
	  indirection->cdr = &scm_undefined;
	  *sterile = c_cons(c_cons(template,
				   indirection),
			    *sterile);
	  return indirection;
	}
      }
    }
  } else {
    // we can't substitute this
    return template;
  }
}

// will have issue matching patterns with stuff after ellipses
bool test_pattern_elided(struct scm_thing * pattern,
			 struct scm_thing * literals,
			 struct scm_thing * form,
			 struct scm_thing * environment,
			 struct scm_thing * top_level_environment) {
  if (form && form->type != PAIR) {
    return false;
  }
  while (form) {
    if (form->type != PAIR) {
      error();
    }
    if (!test_pattern(pattern,
		      literals,
		      form->car,
		      environment,
		      top_level_environment)) {
      return false;
    }
    form = form->cdr;
  }
  return true;
}

// we need to test first because we don't want to end up allocating
// a huge amount of pairs just to have to free them again
bool test_pattern(struct scm_thing * pattern,
		  struct scm_thing * literals,
		  struct scm_thing * form,
		  struct scm_thing * local_environment,
		  struct scm_thing * top_level_environment) {
  if (!pattern) {
    if (!form) {
      return true;
    } else {
      return false;
    }
  } else if (pattern->type == SYMBOL) {
    if (c_member(pattern, literals)) {
      if (c_equal_p(pattern, get_binding(form,
					 local_environment,
					 top_level_environment))) {
	return true;
      } else {
	// doesn't match
	return false;
      }
    } else {
      // non-literal identifier
      return true;
    }
  } else if (pattern->type == PAIR) {
    if (is_elided(pattern)) {
      return test_pattern_elided(pattern->car,
				 literals,
				 form,
				 local_environment,
				 top_level_environment);
    } else if (!form) {
      // can't match
      return false;
    } else if (form->type != PAIR) {
      // also can't match
      return false;
    } else {
      while (pattern &&
	     pattern->type == PAIR &&
	     !is_elided(pattern) &&
	     form &&
	     form->type == PAIR) {
	if (!test_pattern(pattern->car,
			  literals,
			  form->car,
			  local_environment,
			  top_level_environment)) {
	  return false;
	}
	pattern = pattern->cdr;
	form = form->cdr;
      }
      return test_pattern(pattern,
			  literals,
			  form,
			  local_environment,
			  top_level_environment);
    }
  } else if (pattern->type == VECTOR) {
    if (form->type != VECTOR) {
      // well that doesn't match
      return false;
    } else if (c_equal_p(pattern->vector[pattern->vectorsize - 1], &scm_ellipsis)) {
      // this is a funny pattern
      // IMPLEMENTATION RESTRICTION
      error();
#if 0
      if (pattern->vectorsize - 2 > form->vectorsize) {
	// form is too small
	return false;
      } else {
	for (uint32_t i = 0; i < pattern->vectorsize - 2; i++) {
	  if (!test_pattern(pattern->vector[i],
			    literals,
			    form->vector[i],
			    local_environment,
			    top_level_environment)) {
	    return false;
	  }
	}
	for (uint32_t i = pattern->vectorsize - 2; i < form->vectorsize; i++) {
	  if (!test_pattern(pattern->vector[pattern->vectorsize - 2],
			    literals,
			    form->vector[i],
			    local_environment,
			    top_level_environment)) {
	    return false;
	  }
	}
	return true;
      }
#endif
    } else {
      if (pattern->vectorsize != form->vectorsize) {
	return false;
      }
      for (uint32_t i = 0; i < pattern->vectorsize; i++) {
	if (!test_pattern(pattern->vector[i],
			  literals,
			  form->vector[i],
			  local_environment,
			  top_level_environment)) {
	  return false;
	}
      }
      return true;
    }
  } else {
    if (pattern->type == NONE) {
      printf("uninitialised thing in pattern?\n");
      error();
    }
    // assume datum
    if (c_equal_p(pattern, form)) {
      return true;
    } else {
      return false;
    }
  }
}

// will have issue matching patterns with stuff after ellipses
struct scm_thing * match_pattern_elided(struct scm_thing * pattern_big,
					struct scm_thing * literals,
					struct scm_thing * form,
					struct scm_thing * result_head) {
  struct scm_thing * ellipses = pattern_big->cdr;
  struct scm_thing * pattern = pattern_big->car;
  if (pattern->type == SYMBOL) {
    // same as match_pattern
    if (c_member(pattern, literals)) {
      return result_head;
    } else {
      // non-literal identifier
      return c_cons(c_cons(pattern_big, form),
		    result_head);
    }
  } else if (pattern->type == PAIR) {
    struct scm_thing * new_matches = NULL;
    while (form &&
	   form->type == PAIR) {
      new_matches = match_pattern(pattern,
				  literals,
				  form->car,
				  new_matches);
      form = form->cdr;
    }
    struct scm_thing * identifiers = NULL;
    struct scm_thing * slider = new_matches;
    while (slider &&
	   !c_member(slider->car->car, identifiers)) {
      identifiers = c_cons(slider->car->car, identifiers);
      slider = slider->cdr;
    }
    // new_matches is backwards, but now we can merge them all together
    while (identifiers) {
      struct scm_thing * identifier = identifiers->car;
      struct scm_thing * match = NULL;
      slider = new_matches;
      while (slider) {
	while (slider &&
	       !c_equal_p(slider->car->car, identifier)) {
	  slider = slider->cdr;
	}
	if (!slider) {
	  break;
	}
	match = c_cons(slider->car->cdr, match);
	slider = slider->cdr;
      }
      // while matching the do macro this will create a match with
      // ((step ...) ...), but find_all_identifiers will only pick
      // up step and ...
      result_head = c_cons(c_cons(c_cons(identifier,
					 ellipses),
				  match),
			   result_head);
      struct scm_thing * old_identifiers = identifiers;
      identifiers = identifiers->cdr;
      free(old_identifiers);
    }
    free_alist_structure(new_matches);
    return result_head;
  } else if (pattern->type == VECTOR) {
    // implementation restriction
    error();
  } else {
    // datum???
    return result_head;
  }
}

// this will make the list backwards (it is far more convenient) but that shouldn't matter
// make sure that it works first by using test_pattern
struct scm_thing * match_pattern(struct scm_thing * pattern,
				 struct scm_thing * literals,
				 struct scm_thing * form,
				 struct scm_thing * result_head) {
  if (!pattern) {
    return result_head;
  } else if (pattern->type == SYMBOL) {
    if (c_member(pattern, literals)) {
      return result_head;
    } else {
      // non-literal identifier
      return c_cons(c_cons(pattern, form),
		    result_head);
    }
  } else if (pattern->type == PAIR) {
    if (is_elided(pattern)) {
      return match_pattern_elided(pattern,
				  literals,
				  form,
				  result_head);
    } else {
      while (pattern &&
	     pattern->type == PAIR &&
	     !is_elided(pattern) &&
	     form &&
	     form->type == PAIR) {
	result_head = match_pattern(pattern->car,
				    literals,
				    form->car,
				    result_head);
	pattern = pattern->cdr;
	form = form->cdr;
      }
      return match_pattern(pattern,
			   literals,
			   form,
			   result_head);
    }
  } else if (pattern->type == VECTOR) {
    if (c_equal_p(pattern->vector[pattern->vectorsize - 1], &scm_ellipsis)) {
      error();
#if 0
      // this is a funny pattern
      for (uint32_t i = 0; i < pattern->vectorsize - 2; i++) {
	result_head = match_pattern(pattern->vector[i],
				    literals,
				    form->vector[i],
				    result_head);
      }
      if (form->vectorsize > pattern->vectorsize - 2) {
	// note: this doesn't really make sense
	struct scm_thing * new_vector = malloc(sizeof(struct scm_thing));
	new_vector->type = VECTOR;
	new_vector->mutable = true;
	new_vector->vectorsize = form->vectorsize - (pattern->vectorsize - 2);
	new_vector->vector = malloc(sizeof(struct scm_thing *) * new_vector->vectorsize);
	memcpy(new_vector->vector,
	       form->vector + (pattern->vectorsize - 2),
	       new_vector->vectorsize * sizeof(struct scm_thing *));
	return c_cons(c_cons(pattern->vector[pattern->vectorsize - 2],
			     new_vector),
		      result_head);
      } else {
	return result_head;
      }
#endif
    } else {
      for (uint32_t i = 0; i < pattern->vectorsize; i++) {
	result_head = match_pattern(pattern->vector[i],
				    literals,
				    form->vector[i],
				    result_head);
      }
      return result_head;
    }
  } else {
    // assume datum
    return result_head;
  }
}

static struct scm_thing * macro_substitute(struct scm_thing * expression,
					   struct scm_thing * local_environment,
					   struct scm_thing * top_level_environment,
					   struct scm_thing * literals) {
  if (!expression) {
    return NULL;
  }
  if (expression->type == SYMBOL) {
    if (c_member(expression, literals)) {
      // we don't expand the literals
      return expression;
    } else {
      struct scm_thing * value = c_assoc(expression, local_environment);
      if (value) {
	struct scm_thing * rv = malloc(sizeof(struct scm_thing));
	rv->type = INDIRECTION;
	rv->car = value;
	rv->cdr = NULL;
	return rv;
      } else {
	value = c_assoc(expression, top_level_environment);
	if (value) {
	  return value->cdr;
	} else {
	  // in a compiler i would NOT create a binding with the value <undefined>
	  return expression;
	}
      }
    }
  } else if (expression->type == PAIR) {
    struct scm_thing * result_head = malloc(sizeof(struct scm_thing));
    struct scm_thing * result_tail = result_head;
    result_tail->type = PAIR;
    result_tail->car = macro_substitute(expression->car,
					local_environment,
					top_level_environment,
					literals);
    expression = expression->cdr;
    while (expression && expression->type == PAIR) {
      result_tail->cdr = malloc(sizeof(struct scm_thing));
      result_tail = result_tail->cdr;
      result_tail->type = PAIR;
      result_tail->car = macro_substitute(expression->car,
					  local_environment,
					  top_level_environment,
					  literals);
      expression = expression->cdr;
    }
    result_tail->cdr = macro_substitute(expression,
					local_environment,
					top_level_environment,
					literals);
    return result_head;
  } else {
    // not doing vectors?
    return expression;
  }
}

struct scm_thing * scm_syntax_rules(struct scm_thing * args,
				    struct scm_thing * local_environment,
				    struct scm_thing * top_level_environment) {
  if (c_length(args) < 2) {
    error();
  }
  struct scm_thing * literals = args->car;
  struct scm_thing * syntax_rules = args->cdr;
  struct scm_thing * rv = malloc(sizeof(struct scm_thing));
  rv->type = INTERPRETED_MACRO;
  rv->mutable = true;
  rv->car = literals;
  // "complicated" stuff for hygenic macros
  struct scm_thing * result_head = NULL;
  struct scm_thing * result_tail = NULL;
  while (syntax_rules) {
    if (result_tail) {
      result_tail->cdr = malloc(sizeof(struct scm_thing));
      result_tail = result_tail->cdr;
    } else {
      result_head = malloc(sizeof(struct scm_thing));
      result_tail = result_head;
    }
    // i will assume no-one would put ellipses in
    // the template but not in the pattern
    struct scm_thing * pattern = syntax_rules->car->car;
    struct scm_thing * template = syntax_rules->car->cdr;
    result_tail->car = c_cons(pattern,
			      macro_substitute(template,
					       local_environment,
					       top_level_environment,
					       find_all_identifiers(pattern,
								    literals)));
    syntax_rules = syntax_rules->cdr;
  }
  rv->cdr = result_head;
  return rv;
}
  
