#include <string.h>
#include "main.h"
#include "misc.h"
#include "number.h"
#include "character.h"
#include "string.h"
#include "unistd.h"

#include "program.h"
#include "library.h"


const struct scm_thing scm_true = {.type = BOOLEAN, .mutable = false, .boolean = true};
const struct scm_thing scm_false = {.type = BOOLEAN, .mutable = false, .boolean = false};

const struct scm_thing scm_ellipsis = {.type = SYMBOL, .mutable = false, .symbol = "..."};
const struct scm_thing scm_ellipsis_pair = {.type = PAIR, .mutable = false, .car = &scm_ellipsis, .cdr = NULL};

const struct scm_thing scm_undefined = {.type = UNDEFINED, .mutable = false};
const struct scm_thing scm_undefined_symbol = {.type = SYMBOL, .mutable = false, .symbol = "<undefined>"};

const struct scm_thing scm_zero = {
  .type = REAL_RATIONAL,
  .mutable = false,
  .realnumerator = 0,
  .realdenominator = 1
};
const struct scm_thing scm_one = {
  .type = REAL_RATIONAL,
  .mutable = false,
  .realnumerator = 1,
  .realdenominator = 1
};


bool charisinitial(char character) {
  bool rv = false;
  rv |= character >= 'A' && character <= 'Z';
  rv |= character >= 'a' && character <= 'z';
  rv |= character == '!';
  rv |= character >= '$' && character <= '&';
  rv |= character == '*';
  rv |= character == '/';
  rv |= character == ':';
  rv |= character >= '<' && character <= '?';
  rv |= character >= '^' && character <= '_';
  rv |= character == '~';
  return rv;
}

bool charissubsequent(char character) {
  bool rv = charisinitial(character);
  rv |= character >= '0' && character <= '9';
  rv |= character == '+';
  rv |= character == '-';
  rv |= character == '.';
  rv |= character == '@';
  return rv;
}

bool charsarebool(char* characters) {
  if (characters[0] != '#')
    return false;
  if ((characters[1] | 0x20 ) == 't' || (characters[1] | 0x20 ) == 'f')
    return true;
  return false;
}

bool chariswhitespace(char character) {
  switch (character) {
  case ' ':
  case '\n':
  case '\t':
    return true;
  default:
    return false;
  }
}

const struct scm_thing quote_symbol = {
  .type = SYMBOL,
  .mutable = false,
  .symbol = "quote"
};

// (SHOULD) leaves **current_letter on delimiter
struct scm_thing* digestsymbol(char** current_letter) {
  // this should digest a single symbol
  struct scm_thing* symbol = malloc(sizeof(struct scm_thing));
  if (!symbol) {
    printf("couldn't allocate memory?\n");
  }
  symbol->type = NONE;
  // quote
  if ((*current_letter)[0] == '\'') {
    //printf("quote\n");
    (*current_letter)++;
    symbol->type = PAIR;
    symbol->car = &quote_symbol;
    symbol->cdr = malloc(sizeof(struct scm_thing));
    symbol->cdr->type = PAIR;
    symbol->cdr->cdr = NULL;
    if ((*current_letter)[0] == '(') {
	(*current_letter)++;
      if ((*current_letter)[0] == ')') {
	(*current_letter)++;
	// we have a list containing the empty list because this is (quote ()) not (quote)
	symbol->cdr->car = NULL;
	return symbol;
      } else {
	symbol->cdr->car = malloc(sizeof(struct scm_thing));
	symbol->cdr->car->type = PAIR;
	symbol->cdr->car->cdr = NULL;
	return processlist(current_letter, symbol, symbol->cdr->car);
      }
    } else {
      symbol->cdr->car = digestsymbol(current_letter);
      return symbol;
    }
  }
  // identifier
  if (charisinitial((*current_letter)[0])) {
    char* beginning_of_word = *current_letter;
    // printf("initial: %c\n", current_letter[0]);
    while(charissubsequent((*current_letter)[1])) {
      /* printf("subsequent: %c\n", (*current_letter)[1]); */
      (*current_letter)++;
    }
    size_t length = *current_letter - beginning_of_word + 1;
    //printf("identifier: %.*s\n", (int)length, beginning_of_word);
    symbol->type = SYMBOL;
    symbol->symbol = malloc(length + 1);
    memcpy(symbol->symbol, beginning_of_word, length);
    symbol->symbol[length] = '\0';
    (*current_letter)++;
    return symbol;
  }
  if (((*current_letter)[0] == '+' || (*current_letter)[0] == '-') && charisdelimiter((*current_letter)[1])) {
    //printf("perculiar identifier: %c\n", (*current_letter)[0]);
    symbol->type = SYMBOL;
    symbol->symbol = malloc(2);
    symbol->symbol[0] = (*current_letter)[0];
    symbol->symbol[1] = '\0';
    (*current_letter)++;
    return symbol;
  }
  if ((*current_letter)[0] == '.' &&
      (*current_letter)[1] == '.' &&
      (*current_letter)[2] == '.' &&
      charisdelimiter((*current_letter)[3])) {
    //printf("perculiar identifier: ...\n");
    symbol->type = SYMBOL;
    symbol->symbol = "...";
    (*current_letter) += 3;
    return symbol;
  }
  // hashtag stuff
  if ((*current_letter)[0] == '#') {
    // boolean
    if (((*current_letter)[1] | 0x20) == 't') {
      //printf("boolean: #t\n");
      symbol->type = BOOLEAN;
      symbol->boolean = true;
      (*current_letter) += 2;
      return symbol;
    }
    else if (((*current_letter)[1] | 0x20) == 'f') {
      //printf("boolean: #f\n");
      symbol->type = BOOLEAN;
      symbol->boolean = false;
      (*current_letter) += 2;
      return symbol;
    }
    // number
    else if (((*current_letter)[1] | 0x20) == 'b') {
      *symbol = processbinary(current_letter);
      if (symbol->type) {
	return symbol;
      }
    }
    else if (((*current_letter)[1] | 0x20) == 'o') {
      printf("implementation restriction violation: octal numbers not implemented\n");
    }
    else if (((*current_letter)[1] | 0x20) == 'd') {
      *symbol = processdecimal(current_letter);
      if (symbol->type) {
	return symbol;
      }
    }
    else if (((*current_letter)[1] | 0x20) == 'x') {
      printf("implementation restriction violation: hexadecimal numbers not implemented\n");
    }
    else if (((*current_letter)[1] | 0x20) == 'i' || ((*current_letter)[1] | 0x20) == 'e') {
      if ((*current_letter)[2] != '#') {
	*symbol = processdecimal(current_letter);
	if (symbol->type) {
	  return symbol;
	}
      }
      else if (((*current_letter)[3] | 0x20) == 'b') {
	*symbol = processbinary(current_letter);
	if (symbol->type) {
	  return symbol;
	}
      }
      else if (((*current_letter)[3] | 0x20) == 'o') {
	printf("implementation restriction violation: octal numbers not implemented\n");
      }
      else if (((*current_letter)[3] | 0x20) == 'd') {
	*symbol = processdecimal(current_letter);
	if (symbol->type) {
	  return symbol;
	}
      }
      else if (((*current_letter)[3] | 0x20) == 'x') {
	printf("implementation restriction violation: hexadecimal numbers not implemented\n");
      }
    }
    // character
    else if ((*current_letter)[1] == '\\') {
      *symbol = processcharacter(current_letter);
      if (symbol->type) {
	return symbol;
      }
    }
  }
  // decimal numbers not starting with a hashtag
  if ((*current_letter)[0] == '+' ||
      (*current_letter)[0] == '-' ||
      ((*current_letter)[0] >= '0' && (*current_letter)[0] <= '9')) {
    *symbol = processdecimal(current_letter);
    if (symbol->type) {
      return symbol;
    }
  }
  // string
  if ((*current_letter)[0] == '"') {
    *symbol = processstring(current_letter);
    if (symbol->type) {
      return symbol;
    }
  }
  char* beginning_of_word = *current_letter;
  while ((*current_letter)[0] && !charisdelimiter((*current_letter)[0])) {
    (*current_letter)++;
  }
  printf("unknown symbol: %.*s\n", (int)(*current_letter - beginning_of_word + 1), beginning_of_word);
  error();
  return symbol;
}

void skip_insignificant(char ** current_letter) {
  while (chariswhitespace((*current_letter)[0]) ||
	 ((*current_letter)[0] == ';' &&
	  (*current_letter)[1] == ';')) {
    while (chariswhitespace((*current_letter)[0])) {
      (*current_letter)++;
    }
    if ((*current_letter)[0] == ';' &&
	(*current_letter)[1] == ';') {
      while ((*current_letter)[0] != '\n' &&
	     (*current_letter)[0] != '\0') {
	(*current_letter)++;
      }
    }
  }
}
  

struct scm_thing* processlist(char** current_letter, struct scm_thing* firstpair, struct scm_thing* currentpair) {
  // skip to content
  skip_insignificant(current_letter);

  currentpair->type = PAIR;
  if ((*current_letter)[0] == ')') {
    // should not be called on an empty list because that means currentpair has been malloc'd
    error();
  }
  if ((*current_letter)[0] == '(') {
    (*current_letter)++;
    // need special case for null list
    skip_insignificant(current_letter);
    if (!(*current_letter)[0] || (*current_letter)[0] == ')') {
      currentpair->car = NULL;
      (*current_letter)++;
    } else {
      struct scm_thing* newpair = malloc(sizeof(struct scm_thing));
      currentpair->car = processlist(current_letter, newpair, newpair);
    }
  } else if ((*current_letter)[0] != ' ' && (*current_letter)[0] != '\n') {
    //printf("current letter: %c\n", (*current_letter)[0]);
    currentpair->car = digestsymbol(current_letter);
  }

  skip_insignificant(current_letter);

  if ((*current_letter)[0] == '.' && (*current_letter)[1] != '.') {
    (*current_letter)++;
    skip_insignificant(current_letter);
    currentpair->cdr = digestsymbol(current_letter);
    skip_insignificant(current_letter);
    if (*(current_letter)[0] != ')') {
      error();
    }
    (*current_letter)++;
    return firstpair;
  } else if ((*current_letter)[0] == '\0' ||
	     (*current_letter)[0] == ')') {
    currentpair->cdr = NULL;
    (*current_letter)++;
    return firstpair;
  } else {
    struct scm_thing* nextpair = malloc(sizeof(struct scm_thing));
    currentpair->cdr = nextpair;
    return processlist(current_letter, firstpair, nextpair);
  }
}


void printpair(struct scm_thing pair, int depth, int indent, bool start);
void printvector(struct scm_thing thing, int indent);

void printthing(struct scm_thing thing, int indent) {
  switch (thing.type) {
  case NONE:
    printf("untyped thing");
    break;
  case PAIR:
    return printpair(thing, 0, indent, true);
    break;
  case BOOLEAN:
    printf("boolean: %s", thing.boolean ? "true" : "false");
    break;
  case SYMBOL:
    printf("symbol: %s", thing.symbol);
    break;
  case REAL_RATIONAL:
    printf("real rational: %i/%u", thing.realnumerator, thing.realdenominator);
    break;
  case COMPLEX_RATIONAL:
    printf("complex rational: %i/%u+%i/%ui", thing.realnumerator, thing.realdenominator,
	   thing.imaginarynumerator, thing.imaginarydenominator);
    break;
  case REAL_FLOAT:
    printf("real float: %g", thing.realfloat);
    break;
  case COMPLEX_FLOAT:
    printf("complex float: %g+%gi", thing.realfloat, thing.imaginaryfloat);
    break;
  case CHARACTER:
    printf("character: %02x", thing.character);
    break;
  case STRING:
    printf("string: %s", thing.string);
    break;
  case INTERPRETED_PROCEDURE:
    printf("interpeted procedure:\n");
    for (int i = 0; i <= indent; i++) {
      putchar(' ');
    }
    return printpair(thing, 0, indent, true);
    break;
  case COMPILED_PROCEDURE:
    printf("compiled procedure");
    break;
  case COMPILED_MACRO:
    printf("compiled macro");
    break;
  case INDIRECTION:
    printf("indirection: ");
    return printthing(*thing.car->car, indent);
  case VECTOR:
    return printvector(thing, indent);
  default:
    printf("not supported by printthing");
  }
}

void printpair(struct scm_thing pair, int depth, int indent, bool start) {
  if (start) {
    putchar('(');
  } else {
    for (int i = 0; i <= indent; i++) {
      putchar(' ');
    }
  }

  if (!pair.car) {
    printf("()");
  } else {
    printthing(*(pair.car), indent + 1);
  }
  
  if (pair.cdr && pair.cdr->type == PAIR) {
    putchar('\n');
    return printpair(*(pair.cdr), depth, indent, false);
  } else if (pair.cdr) {
    printf(" . ");
    printthing(*(pair.cdr), indent + 3);
  } else {
    // print nothing
  }

  for (int i = 0; i <= depth; i++) {
    putchar(')');
  }
}

void printvector(struct scm_thing vector, int indent) {
  printf("#(");
  for (uint32_t i = 0; i < vector.vectorsize - 1; i++) {
    if (vector.vector[i]) {
      printthing(*vector.vector[i], indent + 2);
    } else {
      printf("()");
    }
    putchar('\n');
    for (int i = 0; i < indent + 2; i++) {
      putchar(' ');
    }
  }
  if (vector.vector[vector.vectorsize - 1]) {
    printthing(*vector.vector[vector.vectorsize - 1], indent + 2);
  } else {
    printf("()");
  } 
  printf(")");
}

int main(const int argc, char** argv) {
  char* current_line;
  char* current_letter;
  size_t linesize;

  current_input_port = malloc(sizeof(struct scm_thing));
  current_input_port->type = INPUT_PORT;
  current_input_port->portfd = STDIN_FILENO;
  current_output_port = malloc(sizeof(struct scm_thing));
  current_output_port->type = OUTPUT_PORT;
  current_output_port->portfd = STDOUT_FILENO;

  getline(&current_line, &linesize, stdin);

  current_letter = current_line;

  struct scm_thing rootpair;
  rootpair.type = PAIR;

  processlist(&current_letter, &rootpair, &rootpair);

  printf("input:\n");
  printpair(rootpair, 0, 0, true);
  putchar('\n');

  create_null_variables();
  create_interactive_variables();
  struct scm_thing * result = scm_begin(&rootpair,
					NULL,
					interactive_variables);

  if (result) {
    printf("result:\n");
    printthing(*result, 0);
    putchar('\n');
  }
  
  return 0;
}
