#pragma once

#include <stdio.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>

enum scm_type {
  NONE = 0,
  PAIR,
  BOOLEAN,
  SYMBOL,
  REAL_RATIONAL,
  REAL_FLOAT,
  COMPLEX_RATIONAL,
  COMPLEX_FLOAT,
  CHARACTER,
  STRING,
  VECTOR,
  INPUT_PORT,
  OUTPUT_PORT,
  COMPILED_PROCEDURE,
  INTERPRETED_PROCEDURE,
  COMPILED_MACRO,
  INTERPRETED_MACRO,
  UNDEFINED,
  INDIRECTION
};

struct scm_thing {
  enum scm_type type;
  bool mutable;
  union {
    struct {
      struct scm_thing* car;
      struct scm_thing* cdr;
    };
    bool boolean;
    char* symbol;
    struct {
      int32_t realnumerator;
      uint32_t realdenominator;
      int32_t imaginarynumerator;
      uint32_t imaginarydenominator;
    };
    struct {
      float realfloat;
      float imaginaryfloat;
    };
    unsigned char character;
    struct {
      char * string;
      uint32_t stringsize;
    };
    struct {
      struct scm_thing ** vector;
      uint32_t vectorsize;
    };
    struct scm_thing * (*procedure)(struct scm_thing *);
    struct scm_thing * (*macro)(struct scm_thing *, struct scm_thing *, struct scm_thing *);
    int portfd;
  };
};

extern const struct scm_thing scm_true;
extern const struct scm_thing scm_false;

extern const struct scm_thing scm_ellipsis;
extern const struct scm_thing scm_ellipsis_pair;

extern const struct scm_thing scm_undefined;
extern const struct scm_thing scm_undefined_symbol;

extern const struct scm_thing scm_zero;
extern const struct scm_thing scm_one;

extern struct scm_thing * current_input_port;
extern struct scm_thing * current_output_port;


static inline void error(void) {
  exit(1);
}


struct scm_thing* processlist(char** current_letter,
			      struct scm_thing* firstpair,
			      struct scm_thing* currentpair);
