#include "number.h"
#include "main.h"

bool charisbinary(const char character) {
    bool rv = character == '0' || character == '1';
    rv |= character == '@';
    rv |= character == '+';
    rv |= character == '-';
    rv |= character == 'i';
    rv |= character == '/';
    rv |= character == '#';
    return rv;
}

bool charisdecimal(const char character) {
    bool rv = charisbinary(character);
    rv |= character >= '2' && character <= '9';
    // exponent markers
    rv |= (character >= 'd' && character <= 'f') || character == 's' || character == 'l';
    // radix point
    rv |= character == '.';
    return rv;
}

void binaryhalfcomplex(const char* characters, int* i, int number[4], bool negative[2]) {
    bool thisnegative = false;
    bool complex = false;
    // printf("first character: %c\n", characters[*i]);
    if (characters[*i] == '@') {
        complex = true;
        (*i)++;
    }
    if (characters[*i] == '+') {
        (*i)++;
    }
    else if (characters[*i] == '-') {
        thisnegative = true;
        (*i)++;
    }
    if (characters[*i] == 'i') {
        negative[1] = thisnegative;
        number[2] = 1;
        number[3] = 1;
        (*i)++;
        return;
    }
    int numerator = 0;
    int denominator = 1;
    int j;
    for (j = *i; characters[j] == '0' || characters[j] == '1' || characters[j] == '#'; j++) {}
    int urealend = j;
    int exponent = 0;
    j--;
    for (; j >= *i; j--) {
        /* printf("character: %c, exponent: %i\n", characters[j], exponent); */
        if (characters[j] == '0') {
            exponent++;
        }
        else if (characters[j] == '1') {
            numerator |= 1 << exponent;
            exponent++;
        }
    }
    *i = urealend;
    if (characters[*i] == '/') {
        for (j = *i + 1; characters[j] == '0' || characters[j] == '1' || characters[j] == '#'; j++) {}
        urealend = j;
        exponent = 0;
        j--;
        denominator = 0;
        for (; j >= *i; j--) {
            // printf("character: %c, exponent: %i\n", characters[j], exponent);
            if (characters[j] == '0') {
                exponent++;
            }
            else if (characters[j] == '1') {
                if (exponent > 32)
                    printf("violation of implementation restriction");
                denominator |= 1 << exponent;
                exponent++;
            }
        }
    }
    *i = urealend;
    if (characters[*i] == 'i' || complex) {
        negative[1] = thisnegative;
        number[2] = numerator;
        number[3] = denominator;
        (*i)++;
    }
    else {
        negative[0] = thisnegative;
        number[0] = numerator;
        number[1] = denominator;
    }
}

struct scm_thing processbinary(char** characters) {
    int numlength = 0;
    int numoffset = 0; // index of the first digit of the part of the number known as "complex"
    // this function is given the characters starting from the hashtag of the radix marker
    bool exact = true;
    // real numerator, real denominator, imaginary numerator, imaginary denominator
    int number[] = {0, 1, 0, 1};
    // real sign, complex sign
    bool negative[] = {false, false};
    struct scm_thing symbol;
    symbol.type = NONE;

    assert((*characters)[0] == '#');
    if ((*characters)[1] == 'b') {
        numoffset = 2;
    }
    else if ((*characters)[1] == 'i') {
        exact = false;
        assert((*characters)[2] == '#' && (*characters)[3] == 'b');
        numoffset = 4;
        goto found_exactness;
    }
    else if ((*characters)[1] == 'e') {
        exact = true;
        assert((*characters)[2] == '#' && (*characters)[3] == 'b');
        numoffset = 4;
        goto found_exactness;
    }
    else {
        printf("processbinary function confused by '#%c'", (*characters)[1]);
    }

    if ((*characters)[2] == '#') {
        if ((*characters)[3] == 'i') {
            exact = false;
            numoffset = 4;
            goto found_exactness;
        }
        if ((*characters)[3] == 'e'){
            exact = true;
            numoffset = 4;
            goto found_exactness;
        }
        printf("processbinary function confused by '#%c'", (*characters)[3]);
        //(*characters) += numlength - 1;
        return symbol;
    }

    // we don't know if the number should be exact or inexact
    // we'll just assume that the programmer hasn't tried to use a decimal point or exponent
    for (int i = numoffset; charisbinary((*characters)[i]); i++) {
        if ((*characters)[i] == '#') {
            exact = false;
            goto found_exactness;
        }
    }

found_exactness:
    for (int i = numoffset; charisbinary((*characters)[i]);) {
        if ((*characters)[i] == '/' || (*characters)[i] == '#') {
            printf("processbinary is confused by %c\n", (*characters)[i]);
            i++;
        }
        else {
            binaryhalfcomplex((*characters), &i, number, negative);
            numlength = i;
        }
    }

    /* printf("binary number (octal representation): #o#%c%c%o/%o%c%o/%oi\n", */
    /*        exact ? 'e' : 'i', */
    /*        negative[0] ? '-' : '+', number[0], number[1], */
    /*        negative[1] ? '-' : '+', number[2], number[3] */
    /*        ); */

    (*characters) += numlength;
    if (exact) {
        symbol.type = REAL_RATIONAL;
        symbol.realnumerator = number[0];
        symbol.realdenominator = number[1];
        symbol.imaginarynumerator = number[2];
        symbol.imaginarydenominator = number[3];
    } else {
        symbol.type = REAL_FLOAT;
        symbol.realfloat = (float)(number[0]) / (float)(number[1]);
        symbol.imaginaryfloat = (float)(number[2]) / (float)(number[3]);
    }
    return symbol;
}

int decimalureal(const char* characters, const int offset, bool* exact, int* numerator, int* denominator, float* flonum) {
    int ureallength = 0;
    int exponent = 0;
    int i = offset;
    bool fraction = false;
    for (; charisdecimal(characters[i]); i++) {
        if (characters[i] == '/') {
            fraction = true;
            break;
        }
        if (characters[i] == 'i' || characters[i] == '+' || characters[i] == '-') {
            break;
        }
        if ((characters[i] >= 'd' && characters[i] <= 'f') || characters[i] == 's' || characters[i] == 'l') {
            break;
        }
        if (characters[i] == '.')
            exponent = i - offset;
    }
    if (exponent == 0 && characters[offset] != '.')
        exponent = i - offset;
    // printf("exponent: %i\n", exponent);
    if (*exact && exponent > 8) {
        printf("implementation restriction violation: exact decimal too large\n");
        *exact = false;
    }
    if (exponent > 8 || exponent != i - offset) {
        if (fraction) {
            printf("i can't be bothered implementing this, represent your number some other way\n");
            return 0;
        }
        for (int j = offset; j < i; j++) {
            exponent--;
            *flonum += (characters[j] - 0x30) * powf(10, exponent);
        }
    }
    for (int j = offset; j < i; j++) {
        exponent--;
        *numerator += (characters[j] - 0x30) * powf(10, exponent);
    }
    if (fraction) {
        i++;
        exponent = 0;
        int denominatorstart = i;
        for (; charisdecimal(characters[i]); i++) {
            exponent++;
            if (characters[i] == 'i' || characters[i] == '+' || characters[i] == '-') {
                break;
            }
            if ((characters[i] >= 'd' && characters[i] <= 'f') || characters[i] == 's' || characters[i] == 'l' || characters[i] == '/' || characters[i] == '.') {
                printf("this fraction is nonsense\n");
                break;
            }
        }
        for (int j = denominatorstart; j < i; j++) {
            exponent--;
            *numerator += (characters[j] - 0x30) * powf(10, exponent);
        }
    }
    ureallength = i - offset;
    return ureallength;
}

struct scm_thing processdecimal(char** characters) {
    // the implementation of complex numbers will be broken here,
  // because I only just realised that complex numbers don't really matter here
    int numlength = 0;
    int numoffset = 0;
    bool exact = true;
    // real numerator, real denominator, imaginary numerator, imaginary denominator
    int number[] = {0, 1, 0, 1};
    float flonum[] = {0, 0};
    // real sign, complex sign
    bool negative[] = {false, false};
    // this will be returned
    struct scm_thing symbol;
    symbol.type = NONE;

    if ((*characters)[0] == '#') {
        if ((*characters)[1] == 'd') {
            numoffset = 2;
            if ((*characters)[2] == '#') {
                numoffset = 4;
                if ((*characters)[3] == 'e')
                    goto found_exactness;
                if ((*characters)[3] == 'i') {
                    exact = false;
                    goto found_exactness;
                }
                printf("processdecimal confused by #%c\n", (*characters)[3]);
                return symbol;
            }
        }
        else if ((*characters)[1] == 'e') {
            numoffset = 2;
            if ((*characters)[2] != '#')
                goto found_exactness;
            if ((*characters)[3] == 'd') {
                numoffset = 4;
                goto found_exactness;
            }
            printf("processdecimal confused by #%c\n", (*characters)[3]);
            return symbol;
        }
        else if ((*characters)[1] == 'i') {
            exact = false;
            numoffset = 2;
            if ((*characters)[2] != '#')
                goto found_exactness;
            if ((*characters)[3] == 'd') {
                numoffset = 4;
                goto found_exactness;
            }
            printf("processdecimal confused by #%c\n", (*characters)[3]);
            return symbol;
        }
        else {
            printf("processdecimal confused by first #: #%c\n", (*characters)[1]);
            return symbol;
        }
    }

    for (int i = numoffset; charisdecimal((*characters)[i]); i++) {
        if ((*characters)[i] == '#' ||
	    (*characters)[i] == '.' ||
	    ((*characters)[i] >= 'd' && (*characters)[i] <= 'f') ||
	    (*characters)[i] == 's' ||
	    (*characters)[i] == 'l') {
            exact = false;
            goto found_exactness;
        }
    }

found_exactness:

    if ((*characters)[numoffset + 1] == 'i') {
        switch ((*characters)[numoffset]) {
            case '-':
                negative[1] = true;
            case '+':
                number[2] = 1;
                numlength = numoffset + 2;
                goto end;
            default:
                printf("processdecimal confused by weird complex thing, did you forget the sign?");
                return symbol;
        }
    }
    if ((*characters)[numoffset] == '-') {
        numoffset++;
        numoffset += decimalureal((*characters), numoffset, &exact, &number[0], &number[1], &flonum[0]);
        if ((*characters)[numoffset] == 'i') {
            negative[1] = true;
            number[2] = number[0];
            number[3] = number[1];
            number[0] = 0;
            number[1] = 1;
            flonum[1] = flonum[0];
            flonum[0] = 0;
            goto end;
        }
        negative[0] = true;
    }
    else {
        if ((*characters)[numoffset] == '+')
            numoffset++;
        // printf("numoffset: %i\n", numoffset);
        numoffset += decimalureal((*characters), numoffset, &exact, &number[0], &number[1], &flonum[0]);
        if ((*characters)[numoffset] == 'i') {
            printf("violation of implementation restriction: imaginary numbers not properly implemented for decimals\n");
            return symbol;
        }
        if ((*characters)[numoffset] == '-' || (*characters)[numoffset] == '+') {
            printf("violation of implementation restriction: complex numbers not properly implemented for decimals\n");
            return symbol;
        }
        if (!charisdelimiter((*characters)[numoffset])) {
            printf("processdecimal confused by %c at end of number\n", (*characters)[numoffset]);
            return symbol;
        }
    }

    end:

    if (flonum[0]) {
      //printf("decimal flonum: #d#%c%e\n", exact ? 'e' : 'i', flonum[0]);
        symbol.type = REAL_FLOAT;
        symbol.realfloat = flonum[0];
    } else if (number[2]) {
      /* printf("decimal number (octal representation): #o#%c%c%o/%o%c%o/%oi\n",
               exact ? 'e' : 'i',
               negative[0] ? '-' : '+', number[0], number[1],
               negative[1] ? '-' : '+', number[2], number[3]
               );*/
        symbol.type = COMPLEX_RATIONAL;
        symbol.realnumerator = negative[0] ? -number[0] : number[0];
        symbol.realdenominator = number[1];
        symbol.imaginarynumerator = negative[1] ? -number[2] : number[2];
        symbol.imaginarydenominator = number[3];
    } else {
      /*printf("decimal number (octal representation): #o#%c%c%o/%o\n",
               exact ? 'e' : 'i',
               negative[0] ? '-' : '+', number[0], number[1]
               );*/
        symbol.type = REAL_RATIONAL;
        symbol.realnumerator = negative[0] ? -number[0] : number[0];
        symbol.realdenominator = number[1];
    }

    numlength = numoffset;
    (*characters) += numlength;
    return symbol;
}
