#include "primitive.h"
#include "eval.h"
#include "standard.h"
#include "program.h"

// this is the simplest macro like, ever
struct scm_thing * scm_quote(struct scm_thing * args,
			     struct scm_thing * local_environment,
			     struct scm_thing * top_level_environment) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * datum = args->car;
  if (datum && datum->type == INDIRECTION) {
    if (!datum->cdr) {
      // i think this is correct
      return datum->car->car;
    } else {
      // no idea if this can happen
      printf("you made something weird happen, now add a test case!\n");
      error();
    }
  }
  return datum;
}

static struct scm_thing * substitute(struct scm_thing * expression,
				     struct scm_thing * local_environment,
				     struct scm_thing * top_level_environment,
				     struct scm_thing * formals) {
  if (!expression) {
    return NULL;
  }
  if (expression->type == INDIRECTION) {
    // this seems kinda dodgy, but it should work
    // what it does is fish the name of the variable out and
    // check if it is in formals
    // we don't do this for indirections generated by macros
    // as their reason to be is to avoid this
    if (!expression->cdr &&
	c_memv(expression->car->car, formals)) {
      //printf("dodgy codepath triggered\n");
      return expression->car->car;
    } else {
      return expression;
    }
  } else if (expression->type == SYMBOL) {
    if (c_member(expression, formals)) {
      // we don't expand the formals
      return expression;
    } else {
      struct scm_thing * value = c_assoc(expression, local_environment);
      if (value) {
	struct scm_thing * rv = malloc(sizeof(struct scm_thing));
	rv->type = INDIRECTION;
	rv->car = value;
	rv->cdr = NULL;
	return rv;
      } else {
	value = c_assoc(expression, top_level_environment);
	if (value) {
	  struct scm_thing * rv = malloc(sizeof(struct scm_thing));
	  rv->type = INDIRECTION;
	  rv->car = value;
	  rv->cdr = NULL;
	  return rv;
	  //return value->cdr;
	} else {
	  // in a compiler i would create a binding with the value <undefined>
	  return expression;
	}
      }
    }
  } else if (expression->type == PAIR) {
    struct scm_thing * result_head = malloc(sizeof(struct scm_thing));
    struct scm_thing * result_tail = result_head;
    result_tail->type = PAIR;
    result_tail->car = substitute(expression->car,
				  local_environment,
				  top_level_environment,
				  formals);
    expression = expression->cdr;
    while (expression && expression->type == PAIR) {
      result_tail->cdr = malloc(sizeof(struct scm_thing));
      result_tail = result_tail->cdr;
      result_tail->type = PAIR;
      result_tail->car = substitute(expression->car,
				    local_environment,
				    top_level_environment,
				    formals);
      expression = expression->cdr;
    }
    result_tail->cdr = substitute(expression,
				  local_environment,
				  top_level_environment,
				  formals);
    return result_head;
  } else {
    // not doing vectors?
    return expression;
  }
}

struct scm_thing * c_lambda(struct scm_thing * formals,
			    struct scm_thing * body,
			    struct scm_thing * local_environment,
			    struct scm_thing * top_level_environment) {
  // all this NONSENSE because we don't store closure
  // ok i can't get it do go through this codepath
  struct scm_thing * slider = formals;
  while (slider) {
    if (slider->car->type == INDIRECTION &&
	!slider->car->cdr) {
      //printf("dodgy codepath 2 triggered\n");
      break;
    } else {
      slider = slider->cdr;
    }
  }
  if (slider) {
    slider = formals;
    struct scm_thing * formals_tail = NULL;
    while (slider) {
      if (formals_tail) {
	formals_tail->cdr = malloc(sizeof(struct scm_thing));
	formals_tail = formals_tail->cdr;
      } else {
	formals = malloc(sizeof(struct scm_thing));
	formals_tail = formals;
      }
      formals_tail->type = PAIR;
      if (slider->car->type == INDIRECTION &&
	  !slider->car->cdr) {
	formals_tail->car = slider->car->car->car;
      } else {
	formals_tail->car = slider->car;
      }
      slider = slider->cdr;
    }
  }
  struct scm_thing * new_body = substitute(body,
					   local_environment,
					   top_level_environment,
					   formals);
  struct scm_thing * rv = malloc(sizeof(struct scm_thing));
  rv->type = INTERPRETED_PROCEDURE;
  rv->mutable = true;
  rv->car = formals;
  rv->cdr = new_body;
  return rv;
}

struct scm_thing * scm_lambda(struct scm_thing * args,
			      struct scm_thing * local_environment,
			      struct scm_thing * top_level_environment) {
  if (c_length(args) < 2) {
    error();
  }
  return c_lambda(args->car,
		  args->cdr,
		  local_environment,
		  top_level_environment);
}
  
struct scm_thing * scm_if(struct scm_thing * args,
			  struct scm_thing * local_environment,
			  struct scm_thing * top_level_environment) {
  if (c_length(args) < 2 || c_length(args) > 3) {
    error();
  }
  struct scm_thing * test = args->car;
  struct scm_thing * consequent = args->cdr->car;
  struct scm_thing * alternate = NULL;
  if (args->cdr->cdr) {
    alternate = args->cdr->cdr->car;
  }
  if (!c_eqv_p(c_eval(test,
		      local_environment,
		      top_level_environment),
	       &scm_false)) {
    return c_eval(consequent,
		  local_environment,
		  top_level_environment);
  } else {
    if (alternate) {
      return c_eval(alternate,
		    local_environment,
		    top_level_environment);
    } else {
      // not an error, but unspecified
      return NULL;
    }
  }
}

struct scm_thing * scm_set(struct scm_thing * args,
			   struct scm_thing * local_environment,
			   struct scm_thing * top_level_environment) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * variable = args->car;
  struct scm_thing * expression = args->cdr->car;
  struct scm_thing * variable_pair;
  if (!variable) {
    error();
  } else if (variable->type == INDIRECTION) {
    variable_pair = variable->car;
    if (variable_pair->cdr && variable_pair->cdr->type == INDIRECTION) {
      error();
    }
  } else {
    variable_pair = c_assoc(variable, local_environment);
    if (!variable_pair) {
      variable_pair = c_assoc(variable, top_level_environment);
      if (!variable_pair) {
	error();
      }
    }
  }
  struct scm_thing * value = c_eval(expression,
				    local_environment,
				    top_level_environment);
  variable_pair->cdr = value;
  return NULL;
}

struct scm_thing * c_procedure_call(struct scm_thing * operator,
				    struct scm_thing * operands,
				    struct scm_thing * local_environment,
				    struct scm_thing * top_level_environment) {
  switch (operator->type) {
  case COMPILED_PROCEDURE:
    return (*operator->procedure)(c_evlis(operands,
					  local_environment,
					  top_level_environment));
  case INTERPRETED_PROCEDURE:
    // the cdr of a lambda call
    struct scm_thing * formals = operator->car;
    struct scm_thing * arguments = c_evlis(operands,
					   local_environment,
					   top_level_environment);
    struct scm_thing * body = operator->cdr;
    // ok, hear me out, procedures have their own closure
    // and so they don't need our local environment
    struct scm_thing * new_local = NULL;
    while (formals &&
	   formals->type == PAIR &&
	   arguments &&
	   arguments->type == PAIR) {
      if (formals->car->type == INDIRECTION) {
	if (formals->car->cdr) {
	  // dodgy thing to get macros to work
	  formals->car->car->cdr = arguments->car;
	} else {
	  // can't deal with this at this point, there's a problem with how we do closure
	  printf("this is how james garfield died\n");
	  error();
	}
      } else {
	new_local = c_cons(c_cons(formals->car,
				  arguments->car),
			   new_local);
      }
      formals = formals->cdr;
      arguments = arguments->cdr;
    }
    if (formals) {
      if (formals->type == PAIR) {
	printf("%i is too few arguments or the arguments weren't a list\n", c_length(operands));
	error();
      } else {
	new_local = c_cons(c_cons(formals,
				  arguments),
			   new_local);
      }
    } else if (arguments) {
      printf("too many arguments\n");
      error();
    }
    // now we have bound our variables we need to eval the body
    // note that the way we bind the variables means that if we call
    // an interpreted procedure that was defined to use a global variable
    // and one of our bound variables shares a name with it then it will be
    // substituted with the bound value rather than the global one
    // this may or may not be r5rs-compliant, but guile doesn't do this
    // note: it would be nice to have a continuation here that would free
    // all of the pairs we just malloc'd
    return scm_begin(body,
		     new_local,
		     top_level_environment);
  default:
    error();
  }
}
