#include "program.h"
#include "eval.h"
#include "standard.h"
#include "primitive.h"

#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>


struct scm_thing * load(struct scm_thing * args);
struct scm_thing * bad_append(struct scm_thing * first_list,
			      struct scm_thing * second_list);

// quirk: cannot end with a define
struct scm_thing * scm_begin(struct scm_thing * args,
			     struct scm_thing * local_environment,
			     struct scm_thing * top_level_environment) {
  if (!args) {
    printf("the empty list is not a valid program!\n");
    // this is nonsense
    error();
  }
  while (args->type == PAIR &&
	 (args->cdr ||
	  (args->car->type == PAIR &&
	   args->car->car->type == SYMBOL &&
	   !strcmp(args->car->car->symbol, "load")))) {
    if (args->car->type == PAIR &&
	args->car->car->type == SYMBOL) {
      if (!strcmp(args->car->car->symbol, "define")) {
	top_level_environment = c_define(args->car->cdr,
					 local_environment,
					 top_level_environment);
	args = args->cdr;
	continue;
      } else if (!strcmp(args->car->car->symbol, "define-syntax")) {
	top_level_environment = c_define_syntax(args->car->cdr,
						local_environment,
						top_level_environment);
	args = args->cdr;
	continue;
      } else if (!strcmp(args->car->car->symbol, "load")) {
	args = bad_append(load(args->car->cdr), args->cdr);
	continue;
      }
    }
    c_eval(args->car,
	   local_environment,
	   top_level_environment);
    args = args->cdr;
  }
  if (args->type != PAIR) {
    // also nonsense
    error();
  }
    
  return c_eval(args->car,
		local_environment,
		top_level_environment);
}


// returns new variable list
struct scm_thing * c_define(struct scm_thing * args,
			    struct scm_thing * local_environment,
			    struct scm_thing * top_level_environment) {
  if (args->car->type == PAIR) {
    if (args->car->car->type != SYMBOL) {
      error();
    }
    struct scm_thing * new_variable = c_lambda(args->car->cdr,
					       args->cdr,
					       local_environment,
					       top_level_environment);
    return c_cons(c_cons(args->car->car, new_variable),
		  top_level_environment);
  } else {
    if (args->car->type != SYMBOL) {
      error();
    }
    if (c_length(args) != 2) {
      error();
    }
    struct scm_thing * new_variable = c_eval(args->cdr->car,
					     local_environment,
					     top_level_environment);
    return c_cons(c_cons(args->car, new_variable),
		  top_level_environment);
  }
}

// returns new variable list
struct scm_thing * c_define_syntax(struct scm_thing * args,
				   struct scm_thing * local_environment,
				   struct scm_thing * top_level_environment) {
  if (args->car->type != SYMBOL) {
    error();
  }
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * new_variable = c_eval(args->cdr->car,
					   local_environment,
					   top_level_environment);
  return c_cons(c_cons(args->car, new_variable),
		top_level_environment);
}

// basically like running read over and over
struct scm_thing * load(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * filename = args->car;
  if (!c_string_p(filename)) {
    error();
  }
  char * buf = malloc((filename->stringsize + 1) * sizeof(char));
  memcpy(buf, filename->string, filename->stringsize);
  buf[filename->stringsize] = '\0';
  int fd = open(buf, O_RDONLY);
  if (fd < 0) {
    perror("failed opening file");
    error();
  }
  struct stat file_stat;
  if (stat(buf, &file_stat) < 0) {
    perror("failed stating file");
    error();
  }
  char * characters = malloc(file_stat.st_size + 1);
  if (read(fd, characters, file_stat.st_size) < 0) {
    perror("failed reading file");
    error();
  }
  characters[file_stat.st_size] = '\0';
  struct scm_thing * root_pair = malloc(sizeof(struct scm_thing));
  root_pair->type = PAIR;
  // we should really have it so that we can free characters, but it will only be a few KB
  processlist(&characters, root_pair, root_pair);
  return root_pair;
}

// will change its first argument
struct scm_thing * bad_append(struct scm_thing * first_list,
			      struct scm_thing * second_list) {
  struct scm_thing * slider = first_list;
  while (slider->cdr) {
    slider = slider->cdr;
  }
  slider->cdr = second_list;
  return first_list;
}
