#pragma once

#include "main.h"

/* 6.1. Equivalence predicates */

bool c_eqv_p(const struct scm_thing * x,
	     const struct scm_thing * y);
struct scm_thing * scm_eqv_p(struct scm_thing * args);

bool c_eq_p(const struct scm_thing * x,
	    const struct scm_thing * y);
struct scm_thing * scm_eq_p(struct scm_thing * args);

bool c_equal_p(const struct scm_thing * x,
	       const struct scm_thing * y);
struct scm_thing * scm_equal_p(struct scm_thing * args);

/* 6.2.5. Numerical operations */

bool c_number_p(struct scm_thing * obj);
struct scm_thing * scm_number_p(struct scm_thing * args);

bool c_complex_p(struct scm_thing * obj);
struct scm_thing * scm_complex_p(struct scm_thing * args);

bool c_real_p(struct scm_thing * obj);
struct scm_thing * scm_real_p(struct scm_thing * args);

bool c_rational_p(struct scm_thing * obj);
struct scm_thing * scm_rational_p(struct scm_thing * args);

bool c_integer_p(struct scm_thing * obj);
struct scm_thing * scm_integer_p(struct scm_thing * args);

bool c_exact_p(struct scm_thing * obj);
struct scm_thing * scm_exact_p(struct scm_thing * args);

bool c_inexact_p(struct scm_thing * obj);
struct scm_thing * scm_inexact_p(struct scm_thing * args);

bool c_numerical_equal(struct scm_thing * z1,
		       struct scm_thing * z2);
struct scm_thing * scm_numerical_equal(struct scm_thing * args);

bool c_numerical_less_than(struct scm_thing * x1,
			   struct scm_thing * x2);
struct scm_thing * scm_numerical_less_than(struct scm_thing * args);

struct scm_thing * scm_numerical_greater_than_or_equal_to(struct scm_thing * args);

bool c_numerical_greater_than(struct scm_thing * x1,
			      struct scm_thing * x2);
struct scm_thing * scm_numerical_greater_than(struct scm_thing * args);

struct scm_thing * scm_numerical_less_than_or_equal_to(struct scm_thing * args);

struct scm_thing * scm_zero_p(struct scm_thing * args);

struct scm_thing * scm_positive_p(struct scm_thing * args);

struct scm_thing * scm_negative_p(struct scm_thing * args);

struct scm_thing * scm_odd_p(struct scm_thing * args);

struct scm_thing * scm_even_p(struct scm_thing * args);

struct scm_thing * c_add(struct scm_thing * z1,
			 struct scm_thing * z2);
struct scm_thing * scm_add(struct scm_thing * args);

struct scm_thing * c_subtract(struct scm_thing * z1,
			      struct scm_thing * z2);
struct scm_thing * scm_subtract(struct scm_thing * args);

struct scm_thing * c_multiply(struct scm_thing * z1,
			      struct scm_thing * z2);
struct scm_thing * scm_multiply(struct scm_thing * args);

struct scm_thing * c_divide(struct scm_thing * z1,
			    struct scm_thing * z2);
struct scm_thing * scm_divide(struct scm_thing * args);

struct scm_thing * scm_abs(struct scm_thing * args);

struct scm_thing * scm_quotient(struct scm_thing * args);

struct scm_thing * scm_remainder(struct scm_thing * args);

struct scm_thing * scm_modulo(struct scm_thing * args);

struct scm_thing * scm_numerator(struct scm_thing * args);

struct scm_thing * scm_denominator(struct scm_thing * args);

struct scm_thing * scm_floor(struct scm_thing * args);

struct scm_thing * scm_ceiling(struct scm_thing * args);

struct scm_thing * scm_truncate(struct scm_thing * args);

struct scm_thing * scm_round(struct scm_thing * args);

/* 6.3. Other data types */

/* 6.3.1. Booleans */

struct scm_thing * scm_not(struct scm_thing * args);

struct scm_thing * scm_boolean_p(struct scm_thing * args);

/* 6.3.2. Pairs and lists */
 
struct scm_thing * scm_pair_p(struct scm_thing * args);
  
struct scm_thing * c_cons(struct scm_thing *x,
			  struct scm_thing *y);
struct scm_thing * scm_cons(struct scm_thing * args);

struct scm_thing * scm_car(struct scm_thing * args);

struct scm_thing * scm_cdr(struct scm_thing * args);

struct scm_thing * scm_set_car(struct scm_thing * args);

struct scm_thing * scm_set_cdr(struct scm_thing * args);

struct scm_thing * scm_null_p(struct scm_thing * args);

struct scm_thing * scm_list_p(struct scm_thing * args);

struct scm_thing * scm_list(struct scm_thing * args);

int32_t c_length(struct scm_thing * list);
struct scm_thing * scm_length(struct scm_thing * args);

struct scm_thing * scm_append(struct scm_thing * args);

struct scm_thing * c_list_tail(struct scm_thing * list,
			       uint32_t k);
struct scm_thing * scm_list_tail(struct scm_thing * args);

struct scm_thing * c_memq(struct scm_thing * obj,
			  struct scm_thing * list);
struct scm_thing * scm_memq(struct scm_thing * args);

struct scm_thing * c_memv(struct scm_thing * obj,
			  struct scm_thing * list);
struct scm_thing * scm_memv(struct scm_thing * args);

struct scm_thing * c_member(struct scm_thing * obj,
			    struct scm_thing * list);
struct scm_thing * scm_member(struct scm_thing * args);

struct scm_thing * scm_assq(struct scm_thing * args);

struct scm_thing * scm_assv(struct scm_thing * args);

struct scm_thing * c_assoc(struct scm_thing *obj,
			   struct scm_thing *alist);
struct scm_thing * scm_assoc(struct scm_thing * args);

/* 6.3.3. Symbols */

struct scm_thing * scm_symbol_p(struct scm_thing * args);

struct scm_thing * scm_symbol_to_string(struct scm_thing * args);

struct scm_thing * scm_string_to_symbol(struct scm_thing * args);

/* 6.3.4. Characters */

bool c_char_p(struct scm_thing * obj);
struct scm_thing * scm_char_p(struct scm_thing * args);

struct scm_thing * scm_char_equals_p(struct scm_thing * args);

struct scm_thing * scm_char_less_than_p(struct scm_thing * args);

struct scm_thing * scm_char_greater_than_p(struct scm_thing * args);

struct scm_thing * scm_char_less_than_equals_p(struct scm_thing * args);

struct scm_thing * scm_char_greater_than_equals_p(struct scm_thing * args);

struct scm_thing * scm_char_ci_equals_p(struct scm_thing * args);

struct scm_thing * scm_char_to_integer(struct scm_thing * args);

struct scm_thing * scm_integer_to_char(struct scm_thing * args);

/* 6.3.5. Strings */

bool c_string_p(struct scm_thing * obj);
struct scm_thing * scm_string_p(struct scm_thing * args);

struct scm_thing * scm_make_string(struct scm_thing * args);

struct scm_thing * scm_string_length(struct scm_thing * args);

struct scm_thing * scm_string_ref(struct scm_thing * args);

struct scm_thing * scm_string_set(struct scm_thing * args);

/* 6.3.6. Vectors */

bool c_vector_p(struct scm_thing * obj);
struct scm_thing * scm_vector_p(struct scm_thing * args);

struct scm_thing * c_make_vector(uint32_t k);
struct scm_thing * scm_make_vector(struct scm_thing * args);

struct scm_thing * scm_vector_length(struct scm_thing * args);

struct scm_thing * scm_vector_ref(struct scm_thing * args);

struct scm_thing * scm_vector_set(struct scm_thing * args);

struct scm_thing * c_vector_fill(struct scm_thing * vector,
				 struct scm_thing * fill);
struct scm_thing * scm_vector_fill(struct scm_thing * args);

/* 6.4. Control features */

bool c_procedure_p(struct scm_thing * obj);
struct scm_thing * scm_procedure_p(struct scm_thing * args);

struct scm_thing * scm_apply(struct scm_thing * args,
			     struct scm_thing * local_environment,
			     struct scm_thing * top_level_environment);

/* 6.5. Eval */

struct scm_thing * scm_eval(struct scm_thing * args);

/* 6.6. Input and output */

/* 6.6.1. Ports */

bool c_input_port_p(struct scm_thing * obj);
struct scm_thing * scm_input_port_p(struct scm_thing * args);

bool c_output_port_p(struct scm_thing * obj);
struct scm_thing * scm_output_port_p(struct scm_thing * args);

struct scm_thing * scm_current_input_port(struct scm_thing * args);

struct scm_thing * scm_current_output_port(struct scm_thing * args);

struct scm_thing * scm_open_input_file(struct scm_thing * args);

struct scm_thing * scm_open_output_file(struct scm_thing * args);

struct scm_thing * scm_close_input_port(struct scm_thing * args);

struct scm_thing * scm_close_output_port(struct scm_thing * args);

/* 6.6.2. Input */

struct scm_thing * scm_read_char(struct scm_thing * args);

struct scm_thing * scm_peek_char(struct scm_thing * args);

bool c_eof_object_p(struct scm_thing * obj);
struct scm_thing * scm_eof_object_p(struct scm_thing * args);

/* 6.6.3. Output */

struct scm_thing * scm_write_char(struct scm_thing * args);

/* 6.6.4. System interface */

/* struct scm_thing * scm_load(struct scm_thing * args); */
