(define-syntax cond
  (syntax-rules (else =>)
    ((cond (else result1 result2 ...))
     (begin result1 result2 ...))
    ((cond (test => result))
     (let ((temp test))
       (if temp (result temp))))
    ((cond (test => result) clause1 clause2 ...)
     (let ((temp test))
       (if temp
	   (result temp)
	   (cond clause1 clause2 ...))))
    ((cond (test)) test)
    ((cond (test) clause1 clause2 ...)
     (let ((temp test))
       (if temp
	   temp
	   (cond clause1 clause2 ...))))
    ((cond (test result1 result2 ...))
     (if test (begin result1 result2 ...)))
    ((cond (test result1 result2 ...)
	   clause1 clause2 ...)
     (if test
	 (begin result1 result2 ...)
	 (cond clause1 clause2 ...)))))

(define-syntax let
  (syntax-rules ()
    ((let ((name val) ...) body1 body2 ...)
     ((lambda (name ...) body1 body2 ...)
      val ...))
    ((let tag ((name val) ...) body1 body2 ...)
     ((letrec ((tag (lambda (name ...)
		      body1 body2 ...)))
	tag)
      val ...))))

(define-syntax letrec
  (syntax-rules ()
    ((letrec ((var1 init1) ...) body ...)
     (letrec "generate temp names"
       (var1 ...)
       ()
       ((var1 init1) ...)
       body ...))
    ((letrec "generate temp names"
       ()
       (temp1 ...)
       ((var1 init1) ...)
       body ...)
     (let ((var1 <undefined>) ...)
       (let ((temp1 init1) ...)
	 (set! var1 temp1)
	 ...
	 body ...)))
    ((letrec "generate temp names"
       (x y ...)
       (temp ...)
       ((var1 init1) ...)
       body ...)
     (letrec "generate temp names"
       (y ...)
       (newtemp temp ...)
       ((var1 init1) ...)
       body ...))))


(define (vector-slice oldvector k n)
  (letrec ((newvector (make-vector n))
	   (internal-loop
	    (lambda (i)
	      (vector-set! newvector
			   i
			   (vector-ref oldvector (+ k i)))
	      (if (< i n)
		  (internal-loop (+ i 1))
		  newvector))))
    (internal-loop 0)))

(define (make-vector-object vector)
  (lambda (operation arg)
    (let ((point 0))
      (cond ((equal? operation 'seek)
	     (set! point arg))
	    ((equal? operation 'read)
	     (set! point (+ arg point))
	     (vector-slice vector
			   (- point arg)
			   arg))))))

(define (get-n-chars port n)
  (letrec ((internal-loop
	    (lambda (a vector)
	      (if (= a n)
		  vector
		  (begin
		    (vector-set! vector
				 a
				 (read-char port))
		    (internal-loop (+ a 1)
				   vector))))))
    (internal-loop 0 (make-vector n))))

(define (match-port-alist port symbols alist)
  (cond ((null? symbols)
	 alist)
	((symbol? (car symbols))
	 (match-port-alist port
			   (cdr symbols)
			   (cons
			    (cons (car symbols)
				  (read-char port))
			    alist)))
	((pair? (car symbols))
	 (match-port-alist port
			   (cdr symbols)
			   (cons
			    (cons (car (car symbols))
				  (get-n-chars port
					       (cdr (car symbols))))
			    alist)))
	(else (begin
		(read-char port)
		(match-port-alist port (cdr symbols) alist)))))

(define (read-elf-header port)
  (match-port-alist
   port
   '(ei-mag0
     ei-mag1
     ei-mag2
     ei-mag3
     ei-class
     ei-data
     ei-version
     ei-osabi
     ei-abiversion
     #f
     #f
     #f
     #f
     #f
     #f
     #f
     (e-type . 2)
     (e-machine . 2)
     (e-version . 4)
     (e-entry . 4)
     (e-phoff . 4)
     (e-shoff . 4)
     (e-flags . 4)
     (e-ehsize . 2)
     (e-phentsize . 2)
     (e-phnum . 2)
     (e-shentsize . 2)
     (e-shnum . 2)
     (e-shstrndx . 2))
   '()))

(define (charvector->u16 charvector)
  (+ (char->integer (vector-ref charvector 0))
     (* 256 (char->integer (vector-ref charvector 1)))))

(define (charvector->u32 charvector)
  (+ (char->integer (vector-ref charvector 0))
     (* 256 (char->integer (vector-ref charvector 1)))
     (* 65536 (char->integer (vector-ref charvector 2)))
     (* 16777216 (char->integer (vector-ref charvector 3)))))

(charvector->u32
 (cdr (assoc
       'e-shoff
       (read-elf-header (open-input-file "hello-world.o")))))
