(load "macros.scm")
(load "library.scm")

(define (for-each-char proc string)
  (do ((i 0 (+ i 1)))
      ((= i (string-length string)))
    (proc (string-ref string i))))

(define (for-each-in-vector proc vector)
  (do ((i 0 (+ i 1)))
      ((= i (vector-length vector)))
    (proc (vector-ref vector i))))

(define (digitize n radix)
  (do ((list (cons (string-ref "0123456789abcdef"
			       (remainder n radix))
		   '())
	     (cons (string-ref "0123456789abcdef"
			       (remainder n radix))
		   list)))
      ((< n radix) list)
    (set! n (quotient n radix))))

(define (number->string z . args)
  (let ((radix (if (null? args)
		   10
		   (car args)))
	(list (if (exact? z)
		  '(#\# #\e)
		  '(#\# #\i))))
    (list->string
     (append list
	     (if (negative? z)
		 '(#\-)
		 '())
	     (cond ((zero? z) '(#\0))
		   ((integer? z)
		    (digitize (abs z) radix))
		   ((real? z)
		    (append (digitize (abs (numerator z)) radix)
			    '(#\/)
			    (digitize (denominator z) radix))))))))

(define (write obj . args)
  (let* ((out-char (if (null? args)
		       write-char
		       (lambda (char) (write-char char (car args)))))
	 (out-string (lambda (string) (for-each-char out-char string)))
	 (out-obj (if (null? args)
		      write
		      (lambda (obj) (write obj (car args))))))
    (cond ((null? obj) (out-string "()"))
	  ((boolean? obj)
	   (if obj
	       (out-string "#t")
	       (out-string "#f")))
	  ((pair? obj)
	   (out-char #\()
	   (do ((pair obj (cdr pair)))
	       ((or (not (pair? pair))
		    (null? (cdr pair)))
		(if (pair? pair)
		    (out-obj (car pair))
		    (begin
		      (out-string ". ")
		      (out-obj pair)))
		(out-char #\)))
	     (out-obj (car pair))
	     (out-char #\space)))
	   ((symbol? obj)
	    ;; good enough for r5rs but not for r7rs
	    (out-string (symbol->string obj)))
	   ((number? obj)
	    (out-string (number->string obj)))
	   ((char? obj)
	    (cond ((eq? obj #\space)
		   (out-string "#\\space"))
		  ((eq? obj #\newline)
		   (out-string "#\\newline"))
		  (else
		   (out-string "#\\")
		   (out-char obj))))
	   ((string? obj)
	    (out-char #\")
	    (for-each-char (lambda (char)
			     (if (or (eq? char #\\)
				     (eq? char #\"))
				 (begin (out-char #\\)
					(out-char char))
				 (out-char char)))
			   obj)
	    (out-char #\"))
	   ((vector? obj)
	    (out-string "#(")
	    (out-obj (vector-ref vector 0))
	    (do ((i 1 (+ i 1)))
		((= i (vector-length vector)))
	      (out-char #\space)
	      (out-obj (vector-ref vector i)))
	    (out-char #\))))
	  ))

(write '(test
	 "test"
	 1
	 2
	 ;;#(3 4)
	 #t
	 #f
	 #\space
	 #\5
	 #\\
	 "\"test\""
	 (6 . 7)
	 (8 9 10)
	 (11 12 . 13)))
